package net.sourceforge.rapidprops;

import static org.objectweb.asm.Opcodes.AASTORE;
import static org.objectweb.asm.Opcodes.ACC_FINAL;
import static org.objectweb.asm.Opcodes.ACC_PUBLIC;
import static org.objectweb.asm.Opcodes.ACC_SUPER;
import static org.objectweb.asm.Opcodes.ACONST_NULL;
import static org.objectweb.asm.Opcodes.ALOAD;
import static org.objectweb.asm.Opcodes.ANEWARRAY;
import static org.objectweb.asm.Opcodes.ARETURN;
import static org.objectweb.asm.Opcodes.ASTORE;
import static org.objectweb.asm.Opcodes.BASTORE;
import static org.objectweb.asm.Opcodes.BIPUSH;
import static org.objectweb.asm.Opcodes.DASTORE;
import static org.objectweb.asm.Opcodes.DRETURN;
import static org.objectweb.asm.Opcodes.DUP;
import static org.objectweb.asm.Opcodes.FASTORE;
import static org.objectweb.asm.Opcodes.FRETURN;
import static org.objectweb.asm.Opcodes.IASTORE;
import static org.objectweb.asm.Opcodes.ICONST_0;
import static org.objectweb.asm.Opcodes.ICONST_1;
import static org.objectweb.asm.Opcodes.INVOKEINTERFACE;
import static org.objectweb.asm.Opcodes.INVOKESPECIAL;
import static org.objectweb.asm.Opcodes.INVOKESTATIC;
import static org.objectweb.asm.Opcodes.IRETURN;
import static org.objectweb.asm.Opcodes.LASTORE;
import static org.objectweb.asm.Opcodes.LRETURN;
import static org.objectweb.asm.Opcodes.NEW;
import static org.objectweb.asm.Opcodes.NEWARRAY;
import static org.objectweb.asm.Opcodes.POP;
import static org.objectweb.asm.Opcodes.RETURN;
import static org.objectweb.asm.Opcodes.SASTORE;
import static org.objectweb.asm.Opcodes.SIPUSH;
import static org.objectweb.asm.Opcodes.T_BOOLEAN;
import static org.objectweb.asm.Opcodes.T_BYTE;
import static org.objectweb.asm.Opcodes.T_DOUBLE;
import static org.objectweb.asm.Opcodes.T_FLOAT;
import static org.objectweb.asm.Opcodes.T_INT;
import static org.objectweb.asm.Opcodes.T_LONG;
import static org.objectweb.asm.Opcodes.T_SHORT;
import static org.objectweb.asm.Opcodes.V1_6;

import java.lang.reflect.Array;
import java.lang.reflect.GenericArrayType;
import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.WildcardType;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.regex.Pattern;

import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Type;

public class PropsHelper<T> {
	private static final String CLASS_NAME_DOT_PREFIX = "net.sourceforge.rapidprops.";
	private static final String CLASS_NAME_SLASH_PREFIX = CLASS_NAME_DOT_PREFIX.replaceAll("\\.", "/");

	private static AtomicInteger CURRENT_CLASS_ID = new AtomicInteger();

	private final int classId = CURRENT_CLASS_ID.getAndIncrement();
	private final String classNameDot;
	private final String classNameSlash;
	
	private final Class<T> propsInterface;
	private final Map<?,?> map;
	
	private final ClassWriter cw = new ClassWriter(ClassWriter.COMPUTE_FRAMES | ClassWriter.COMPUTE_MAXS);
	private MethodVisitor mv;

	public PropsHelper(Class<T> propsInterface, Map<?,?> map) {
		this.propsInterface = propsInterface;
		this.map = map;
		String implName = propsInterface.getSimpleName() + "Impl$" + classId;
		this.classNameDot = CLASS_NAME_DOT_PREFIX + implName;
		this.classNameSlash = CLASS_NAME_SLASH_PREFIX + implName;
	}
	
	public String getClassNameDot() {
		return classNameDot;
	}
	
	public byte[] createByteCode() {
		if(!propsInterface.isInterface()) {
			throw new RapidPropsException("Not an interface: " + propsInterface.getName());
		}

		boolean trimValues = true;
		Props propsAnnotation = propsInterface.getAnnotation(Props.class);
		if(propsAnnotation != null) {
			trimValues = propsAnnotation.trimValues();
		}
		
		startImpl();
		
		for(Method m : propsInterface.getMethods()) {
			String methodName = m.getName();
			if(m.getParameterTypes().length > 0) {
				throw new RapidPropsException("Invalid method: " + methodName + ". Only methods without parameters are allowed.");
			}
			
			mv = cw.visitMethod(ACC_PUBLIC + ACC_FINAL, m.getName(), Type.getMethodDescriptor(m), null, null);		
			mv.visitCode();

			Class<?> returnType = m.getReturnType();

			if(isStringOrPrimitiveWrapper(returnType)) {
				createObjectGetter(m);
			} else if(returnType.equals(byte.class)) {
				createPrimitiveByteGetter(m);
			} else if(returnType.equals(short.class)) {
				createPrimitiveShortGetter(m);
			} else if(returnType.equals(int.class)) {
				createPrimitiveIntGetter(m);
			} else if(returnType.equals(long.class)) {
				createPrimitiveLongGetter(m);
			} else if(returnType.equals(double.class)) {
				createPrimitiveDoubleGetter(m);
			} else if(returnType.equals(float.class)) {
				createPrimitiveFloatGetter(m);
			} else if(returnType.equals(boolean.class)) {
				createPrimitiveBooleanGetter(m);				

			} else if(returnType.equals(List.class)) {
				Class<?> parameterClass = getParameterizedClass(m, 0);
				pushArray(m, parameterClass, trimValues);
				mv.visitMethodInsn(INVOKESTATIC, "java/util/Arrays", "asList", "([Ljava/lang/Object;)Ljava/util/List;");
				mv.visitInsn(ARETURN);
			} else if(returnType.isArray()) {
				pushArray(m, trimValues);
				mv.visitInsn(ARETURN);

			} else if(returnType.equals(Map.class)) {
				pushMap(m, 1, trimValues);
				mv.visitInsn(ARETURN);
			} else {
				throw new RapidPropsException("Unsupported return type for method " + methodName + ": " + returnType.getName());
			}

			mv.visitMaxs(0, 0);
			mv.visitEnd();
			
		}
		
		return cw.toByteArray();
	}

	private void pushMap(Method m, int varIndex, boolean trimValues) {		
		mv.visitTypeInsn(NEW, "java/util/HashMap");
		mv.visitInsn(DUP);
		mv.visitMethodInsn(INVOKESPECIAL, "java/util/HashMap", "<init>", "()V");
		mv.visitVarInsn(ASTORE, varIndex);

		Class<?> keyClass = getParameterizedClass(m, 0);
		Class<?> valueClass = getParameterizedClass(m, 1);
		Map<String, String> stringMap = getStringMap(m);

		Prop prop = m.getAnnotation(Prop.class);
		String keyPrefix = getPropertyKey(m, prop) + ".";
		
		for(Entry<String, String> entry : stringMap.entrySet()) {
			mv.visitVarInsn(ALOAD, varIndex);
			String keyAsString = entry.getKey();
			String valueAsString = entry.getValue();
			pushObject(m, keyPrefix + keyAsString, keyClass, keyAsString, trimValues);
			pushObject(m, keyPrefix + keyAsString, valueClass, valueAsString, trimValues);
			mv.visitMethodInsn(INVOKEINTERFACE, "java/util/Map", "put", "(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;");
			mv.visitInsn(POP);
		}		
		mv.visitVarInsn(ALOAD, varIndex);
	}

	private void pushObject(Method m, String propKey, Class<?> type, String valueAsString, boolean trimValues) {
		if(isStringOrPrimitiveWrapper(type)) {
			pushStringOrPrimitiveWrapper(valueAsString, type);
		} else if(type.isArray()) {
			String[] componentStringValues = getComponentStringValues(m, valueAsString, trimValues);
			pushArray(m, type.getComponentType(), propKey, componentStringValues);
		} else {
			throw new RapidPropsException(m.getName() + ": Unsupported element type: " + type);
		}
	}
	
	private void pushArray(Method m, boolean trimValues) {
		Class<?> componentType = m.getReturnType().getComponentType();
		pushArray(m, componentType, trimValues);
	}

	private void pushArray(Method m, Class<?> componentType, boolean trimValues) {
		Prop prop = m.getAnnotation(Prop.class);
		String propKey = getPropertyKey(m, prop);
		String[] componentStringValues = getComponentStringValues(m, trimValues);
		pushArray(m, componentType, propKey, componentStringValues);
	}

	private void pushArray(Method m, Class<?> componentType, String propKey, String[] componentStringValues) {
		if(componentType.equals(String.class)) {
			pushObjectArray(componentStringValues);
		} else if(componentType.equals(Byte.class)) {
			Byte[] array = toByteArray(propKey, componentStringValues);
			pushObjectArray(array);
		} else if(componentType.equals(Short.class)) {
			Short[] array = toShortArray(propKey, componentStringValues);
			pushObjectArray(array);
		} else if(componentType.equals(Integer.class)) {
			Integer[] array = toIntegerArray(propKey, componentStringValues);
			pushObjectArray(array);
		} else if(componentType.equals(Long.class)) {
			Long[] array = toLongArray(propKey, componentStringValues);
			pushObjectArray(array);
		} else if(componentType.equals(Double.class)) {
			Double[] array = toDoubleArray(propKey, componentStringValues);
			pushObjectArray(array);
		} else if(componentType.equals(Float.class)) {
			Float[] array = toFloatArray(propKey, componentStringValues);
			pushObjectArray(array);
		} else if(componentType.equals(Boolean.class)) {
			Boolean[] array = toBooleanArray(componentStringValues);
			pushObjectArray(array);

		} else if(componentType.equals(byte.class)) {
			byte[] array = toPrimitiveByteArray(propKey, componentStringValues);
			pushPrimitiveByteArray(array);
		} else if(componentType.equals(short.class)) {
			short[] array = toPrimitiveShortArray(propKey, componentStringValues);
			pushPrimitiveShortArray(array);
		} else if(componentType.equals(int.class)) {
			int[] array = toPrimitiveIntArray(propKey, componentStringValues);
			pushPrimitiveIntArray(array);
		} else if(componentType.equals(long.class)) {
			long[] array = toPrimitiveLongArray(propKey, componentStringValues);
			pushPrimitiveLongArray(array);
		} else if(componentType.equals(double.class)) {
			double[] array = toPrimitiveDoubleArray(propKey, componentStringValues);
			pushPrimitiveDoubleArray(array);
		} else if(componentType.equals(float.class)) {
			float[] array = toPrimitiveFloatArray(propKey, componentStringValues);
			pushPrimitiveFloatArray(array);
		} else if(componentType.equals(boolean.class)) {
			boolean[] array = toPrimitiveBooleanArray(propKey, componentStringValues);
			pushPrimitiveBooleanArray(array);
		} else if(componentType.isArray()) {
			throw new RapidPropsException(m.getName() + ": Arrays are not supported as element types.");
		} else {
			throw new RapidPropsException(m.getName() + ": Unsupported element type: " + componentType);
		}
	}

	private String[] getComponentStringValues(Method m, boolean trimValues) {
		String valuesAsString = getStringPropValue(m);
		return getComponentStringValues(m, valuesAsString, trimValues);
	}

	private static String[] getComponentStringValues(Method m, String valuesAsString, boolean trimValues) {
		String delimiter = PropConstants.DEFAULT_DELIMITER;
		String nullAlias = PropConstants.NULL_ALIAS;
		String defaultElementValue = PropConstants.NOT_AVAILABLE;
		Prop prop = m.getAnnotation(Prop.class);
		if(prop != null) {
			delimiter = prop.delimiter();
			nullAlias = prop.nullAlias();
			defaultElementValue = prop.defaultElementValue();
		}
		String[] componentStringValues;
		if(valuesAsString == null) {
			componentStringValues = null;
		} else if (Prop.EMPTY.equals(valuesAsString)) {
			componentStringValues = new String[0];
		} else {
			componentStringValues = valuesAsString.split(Pattern.quote(delimiter), -1);
			if(trimValues) {
				for(int i=0; i<componentStringValues.length; i++) {
					if(componentStringValues[i] != null) {
						componentStringValues[i] = componentStringValues[i].trim();
					}
				}
			}
			if(!PropConstants.NOT_AVAILABLE.equals(defaultElementValue)) {
				for(int i=0; i<componentStringValues.length; i++) {				
					if("".equals(componentStringValues[i])) {
						componentStringValues[i] = defaultElementValue;
					}
				}
			}
			for(int i=0; i<componentStringValues.length; i++) {				
				if(nullAlias.equals(componentStringValues[i])) {
					componentStringValues[i] = null;
				}
			}
		}
		return componentStringValues;
	}
	
	private void startImpl() {
		String interfaceSlashName = propsInterface.getName().replaceAll("\\.", "/");

		cw.visit(V1_6, ACC_PUBLIC + ACC_FINAL + ACC_SUPER, classNameSlash, null, "java/lang/Object", new String[] { interfaceSlashName });

		mv = cw.visitMethod(ACC_PUBLIC, "<init>", "()V", null, null);
		mv.visitCode();
		mv.visitVarInsn(ALOAD, 0);
		mv.visitMethodInsn(INVOKESPECIAL, "java/lang/Object", "<init>", "()V");
		mv.visitInsn(RETURN);
		mv.visitMaxs(0, 0);
		mv.visitEnd();
	}
	
	public static byte[] toPrimitiveByteArray(String propKey, String[] stringValues) {
		if(stringValues == null) return null;
		int len = stringValues.length;
		byte[] byteArray = new byte[len];
		for(int i=0; i<len; i++) {
			try {
				if(stringValues[i] == null) {
					throw new RapidPropsException("Element #" + i + " is null in property '" + propKey);
				}
				byteArray[i] = Byte.parseByte(stringValues[i]);
			} catch(NumberFormatException e) {
				throw new RapidPropsException("Element #" + i + " is not a byte in property '" + propKey + "': " + stringValues[i] , e);
			}
		}
		return byteArray;
	}
	
	public static Byte[] toByteArray(String propKey, String[] stringValues) {
		if(stringValues == null) return null;
		int len = stringValues.length;
		Byte[] byteArray = new Byte[len];
		for(int i=0; i<len; i++) {
			try {
				if(stringValues[i] != null) {
					byteArray[i] = new Byte(stringValues[i]);
				}
			} catch(NumberFormatException e) {
				throw new RapidPropsException("Not a byte value in element #" + i + " of property '" + propKey + "': " + stringValues[i] , e);
			}
		}
		return byteArray;
	}
	
	public static short[] toPrimitiveShortArray(String propKey, String[] stringValues) {
		if(stringValues == null) return null;
		int len = stringValues.length;
		short[] shortArray = new short[len];
		for(int i=0; i<len; i++) {
			try {
				if(stringValues[i] == null) {
					throw new RapidPropsException("Element #" + i + " is null in property '" + propKey);
				}
				shortArray[i] = Short.parseShort(stringValues[i]);
			} catch(NumberFormatException e) {
				throw new RapidPropsException("Not a short value in element #" + i + " of property '" + propKey + "': " + stringValues[i] , e);
			}
		}
		return shortArray;
	}
	
	public static Short[] toShortArray(String propKey, String[] stringValues) {
		if(stringValues == null) return null;
		int len = stringValues.length;
		Short[] shortArray = new Short[len];
		for(int i=0; i<len; i++) {
			try {
				if(stringValues[i] != null) {
					shortArray[i] = new Short(stringValues[i]);
				}
			} catch(NumberFormatException e) {
				throw new RapidPropsException("Not a short value in element #" + i + " of property '" + propKey + "': " + stringValues[i] , e);
			}
		}
		return shortArray;
	}

	public static int[] toPrimitiveIntArray(String propKey, String[] stringValues) {
		if(stringValues == null) return null;
		int len = stringValues.length;
		int[] intArray = new int[len];
		for(int i=0; i<len; i++) {
			try {
				if(stringValues[i] == null) {
					throw new RapidPropsException("Element #" + i + " is null in property '" + propKey);
				}
				intArray[i] = Integer.parseInt(stringValues[i]);
			} catch(NumberFormatException e) {
				throw new RapidPropsException("Not an int value in element #" + i + " of property '" + propKey + "': " + stringValues[i] , e);
			}
		}
		return intArray;
	}
	
	public static Integer[] toIntegerArray(String propKey, String[] stringValues) {
		if(stringValues == null) return null;
		int len = stringValues.length;
		Integer[] integerArray = new Integer[len];
		for(int i=0; i<len; i++) {
			try {
				if(stringValues[i] != null) {
					integerArray[i] = new Integer(stringValues[i]);
				}
			} catch(NumberFormatException e) {
				throw new RapidPropsException("Not an integer value in element #" + i + " of property '" + propKey + "': " + stringValues[i] , e);
			}
		}
		return integerArray;
	}

	public static long[] toPrimitiveLongArray(String propKey, String[] stringValues) {
		if(stringValues == null) return null;
		int len = stringValues.length;
		long[] longArray = new long[len];
		for(int i=0; i<len; i++) {
			try {
				if(stringValues[i] == null) {
					throw new RapidPropsException("Element #" + i + " is null in property '" + propKey);
				}
				longArray[i] = Long.parseLong(stringValues[i]);
			} catch(NumberFormatException e) {
				throw new RapidPropsException("Not a long value in element #" + i + " of property '" + propKey + "': " + stringValues[i] , e);
			}
		}
		return longArray;
	}
	
	public static Long[] toLongArray(String propKey, String[] stringValues) {
		if(stringValues == null) return null;
		int len = stringValues.length;
		Long[] longArray = new Long[len];
		for(int i=0; i<len; i++) {
			try {
				if(stringValues[i] != null) {
					longArray[i] = new Long(stringValues[i]);
				}
			} catch(NumberFormatException e) {
				throw new RapidPropsException("Not a long value in element #" + i + " of property '" + propKey + "': " + stringValues[i] , e);
			}
		}
		return longArray;
	}

	public static double[] toPrimitiveDoubleArray(String propKey, String[] stringValues) {
		if(stringValues == null) return null;
		int len = stringValues.length;
		double[] doubleArray = new double[len];
		for(int i=0; i<len; i++) {
			try {
				if(stringValues[i] == null) {
					throw new RapidPropsException("Element #" + i + " is null in property '" + propKey);
				}
				doubleArray[i] = Double.parseDouble(stringValues[i]);
			} catch(NumberFormatException e) {
				throw new RapidPropsException("Not a double value in element #" + i + " of property '" + propKey + "': " + stringValues[i] , e);
			}
		}
		return doubleArray;
	}
	
	public static Double[] toDoubleArray(String propKey, String[] stringValues) {
		if(stringValues == null) return null;
		int len = stringValues.length;
		Double[] doubleArray = new Double[len];
		for(int i=0; i<len; i++) {
			try {
				if(stringValues[i] != null) {
					doubleArray[i] = new Double(stringValues[i]);
				}
			} catch(NumberFormatException e) {
				throw new RapidPropsException("Not a double value in element #" + i + " of property '" + propKey + "': " + stringValues[i] , e);
			}
		}
		return doubleArray;
	}

	public static float[] toPrimitiveFloatArray(String propKey, String[] stringValues) {
		if(stringValues == null) return null;
		int len = stringValues.length;
		float[] floatArray = new float[len];
		for(int i=0; i<len; i++) {
			try {
				if(stringValues[i] == null) {
					throw new RapidPropsException("Element #" + i + " is null in property '" + propKey);
				}
				floatArray[i] = Float.parseFloat(stringValues[i]);
			} catch(NumberFormatException e) {
				throw new RapidPropsException("Not a float value in element #" + i + " of property '" + propKey + "': " + stringValues[i] , e);
			}
		}
		return floatArray;
	}
	
	public static Float[] toFloatArray(String propKey, String[] stringValues) {
		if(stringValues == null) return null;
		int len = stringValues.length;
		Float[] floatArray = new Float[len];
		for(int i=0; i<len; i++) {
			try {
				if(stringValues[i] != null) {
					floatArray[i] = new Float(stringValues[i]);
				}
			} catch(NumberFormatException e) {
				throw new RapidPropsException("Not a float value in element #" + i + " of property '" + propKey + "': " + stringValues[i] , e);
			}
		}
		return floatArray;
	}

	public static boolean[] toPrimitiveBooleanArray(String propKey, String[] stringValues) {
		if(stringValues == null) return null;
		int len = stringValues.length;
		boolean[] booleanArray = new boolean[len];
		for(int i=0; i<len; i++) {
			if(stringValues[i] == null) {
				throw new RapidPropsException("Element #" + i + " is null in property '" + propKey);
			}
			booleanArray[i] = Boolean.parseBoolean(stringValues[i]);
		}
		return booleanArray;
	}
	
	public static Boolean[] toBooleanArray(String[] stringValues) {
		if(stringValues == null) return null;
		int len = stringValues.length;
		Boolean[] booleanArray = new Boolean[len];
		for(int i=0; i<len; i++) {
			if(stringValues[i] != null) {
				booleanArray[i] = new Boolean(stringValues[i]);
			}
		}
		return booleanArray;
	}


	private String getStringPropValue(Method m) {
		Prop prop = m.getAnnotation(Prop.class);
		String propKey = getPropertyKey(m, prop);
		return getStringPropValue(prop, propKey);
	}

	private String getStringPropValue(Prop prop, String propKey) {
		String value;
		if(map.containsKey(propKey)) {
			Object objVal = map.get(propKey);
			value = (objVal == null) ? null : String.valueOf(objVal);
		} else {
			value = getDefaultStringValue(propKey, prop);
		}
		String nullAlias = PropConstants.NULL_ALIAS;
		if(prop != null) {
			nullAlias = prop.nullAlias();
		}
		if(nullAlias.equals(value)) {
			value = null;
		}
		return value;
	}
	
	private Map<String, String> getStringMap(Method m) {
		Prop prop = m.getAnnotation(Prop.class);
		String keyPrefix = getPropertyKey(m, prop) + ".";
		int keyPrefixLen = keyPrefix.length();
		Map<String, String> stringMap = new HashMap<String, String>();
		for(Object objKey : map.keySet()) {
			if(objKey == null) continue;
			String strKey = objKey.toString();
			if(strKey.startsWith(keyPrefix)) {
				String key = strKey.substring(keyPrefixLen);
				String value = getStringPropValue(prop, strKey);
				stringMap.put(key, value);
			}
		}
		return stringMap;
	}

	private Object getStringOrPrimitiveWrapper(Method m) {
		Class<?> type = m.getReturnType();
		if(!isStringOrPrimitiveWrapper(type)) {
			throw new RapidPropsException("Not a String or primitive wrapper type: " + type.getName());
		}
		Object value = getStringOrPrimitiveWrapper(type, getStringPropValue(m));
		return value;
	}

	private static boolean isStringOrPrimitiveWrapper(Class<?> type) {
		if(type.equals(String.class)) return true;
		if(type.equals(Boolean.class)) return true;
		if(type.equals(Byte.class)) return true;
		if(type.equals(Short.class)) return true;
		if(type.equals(Integer.class)) return true;
		if(type.equals(Long.class)) return true;
		if(type.equals(Double.class)) return true;
		if(type.equals(Float.class)) return true;
		return false;
	}
	
	private static Object getStringOrPrimitiveWrapper(Class<?> type, String value) {
		if(value == null) return null;
		if(type.equals(String.class)) {
			return value;
		} else if(type.equals(Byte.class)) {
			return new Byte(value);
		} else if(type.equals(Short.class)) {
			return new Short(value);
		} else if(type.equals(Integer.class)) {
			return new Integer(value);
		} else if(type.equals(Long.class)) {
			return new Long(value);
		} else if(type.equals(Double.class)) {
			return new Double(value);
		} else if(type.equals(Float.class)) {
			return new Float(value);
		} else if(type.equals(Boolean.class)) {
			return new Boolean(value);
		} else {
			throw new RapidPropsException("Unsupported object type: " + type.getName());
		}
	}
	
	private void pushStringOrPrimitiveWrapper(Object value) {		
		if(value == null) {
			mv.visitInsn(ACONST_NULL);
		} else if(value instanceof String) {
			pushString((String)value);
		} else if(value instanceof Boolean) {
			pushBoolean((Boolean)value);
		} else if(value instanceof Byte) {
			pushByte((Byte)value);
		} else if(value instanceof Short) {
			pushShort((Short)value);
		} else if(value instanceof Integer) {
			pushInteger((Integer)value);
		} else if(value instanceof Long) {
			pushLong((Long)value);
		} else if(value instanceof Double) {
			pushDouble((Double)value);
		} else if(value instanceof Float) {
			pushFloat((Float)value);
		} else {
			throw new RapidPropsException("Unsupported value: " + value);
		}
	}

	private void pushStringOrPrimitiveWrapper(String value, Class<?> type) {
		try {
			if(value == null) {
				mv.visitInsn(ACONST_NULL);
			} else if(String.class.equals(type)) {
				pushString(value);
			} else if(Boolean.class.equals(type)) {
				pushBoolean(Boolean.valueOf(value));
			} else if(Byte.class.equals(type)) {
				pushByte(Byte.valueOf(value));
			} else if(Short.class.equals(type)) {
				pushShort(Short.valueOf(value));
			} else if(Integer.class.equals(type)) {
				pushInteger(Integer.valueOf(value));
			} else if(Long.class.equals(type)) {
				pushLong(Long.valueOf(value));
			} else if(Double.class.equals(type)) {
				pushDouble(Double.valueOf(value));
			} else if(Float.class.equals(type)) {
				pushFloat(Float.valueOf(value));
			} else {
				throw new RapidPropsException("Unsupported value: " + value);
			}
		} catch(NumberFormatException e) {
			throw new RapidPropsException("Invalid " + type.getSimpleName() + " value: " + value);
		}
	}

	private void pushString(String value) {
		mv.visitLdcInsn(value);
	}
	
	private void pushByte(Byte value) {
		mv.visitTypeInsn(NEW, "java/lang/Byte");
		mv.visitInsn(DUP);
		mv.visitIntInsn(BIPUSH, value);
		mv.visitMethodInsn(INVOKESPECIAL, "java/lang/Byte", "<init>", "(B)V");
	}
	
	private void pushShort(Short value) {
		mv.visitTypeInsn(NEW, "java/lang/Short");
		mv.visitInsn(DUP);
		mv.visitIntInsn(SIPUSH, value);
		mv.visitMethodInsn(INVOKESPECIAL, "java/lang/Short", "<init>", "(S)V");
	}
	
	private void pushInteger(Integer value) {
		mv.visitTypeInsn(NEW, "java/lang/Integer");
		mv.visitInsn(DUP);
		mv.visitLdcInsn(value);
		mv.visitMethodInsn(INVOKESPECIAL, "java/lang/Integer", "<init>", "(I)V");
	}
	
	private void pushLong(Long value) {
		mv.visitTypeInsn(NEW, "java/lang/Long");
		mv.visitInsn(DUP);
		mv.visitLdcInsn(value);
		mv.visitMethodInsn(INVOKESPECIAL, "java/lang/Long", "<init>", "(J)V");
	}
	
	private void pushDouble(Double value) {
		mv.visitTypeInsn(NEW, "java/lang/Double");
		mv.visitInsn(DUP);
		mv.visitLdcInsn(value);
		mv.visitMethodInsn(INVOKESPECIAL, "java/lang/Double", "<init>", "(D)V");
	}
	
	private void pushFloat(Float value) {
		mv.visitTypeInsn(NEW, "java/lang/Float");
		mv.visitInsn(DUP);
		mv.visitLdcInsn(value);
		mv.visitMethodInsn(INVOKESPECIAL, "java/lang/Float", "<init>", "(F)V");
	}
	
	private void pushBoolean(Boolean value) {
		mv.visitTypeInsn(NEW, "java/lang/Boolean");
		mv.visitInsn(DUP);
		mv.visitInsn(Boolean.TRUE.equals(value) ? ICONST_1 : ICONST_0);
		mv.visitMethodInsn(INVOKESPECIAL, "java/lang/Boolean", "<init>", "(Z)V");
	}
	
	private void pushPrimitiveByteArray(byte[] values) {
		mv.visitIntInsn(SIPUSH, values.length);
		mv.visitIntInsn(NEWARRAY, T_BYTE);
		for(int i=0; i<values.length; i++) {
			mv.visitInsn(DUP);
			mv.visitIntInsn(SIPUSH, i);
			mv.visitIntInsn(BIPUSH, values[i]);
			mv.visitInsn(BASTORE);
		}
	}
	
	private void pushPrimitiveShortArray(short[] values) {
		mv.visitIntInsn(SIPUSH, values.length);
		mv.visitIntInsn(NEWARRAY, T_SHORT);
		for(int i=0; i<values.length; i++) {
			mv.visitInsn(DUP);
			mv.visitIntInsn(SIPUSH, i);
			mv.visitIntInsn(BIPUSH, values[i]);
			mv.visitInsn(SASTORE);
		}
	}
	
	private void pushPrimitiveIntArray(int[] values) {
		mv.visitIntInsn(SIPUSH, values.length);
		mv.visitIntInsn(NEWARRAY, T_INT);
		for(int i=0; i<values.length; i++) {
			mv.visitInsn(DUP);
			mv.visitIntInsn(SIPUSH, i);
			mv.visitIntInsn(SIPUSH, values[i]);
			mv.visitInsn(IASTORE);
		}
	}
	
	private void pushPrimitiveLongArray(long[] values) {
		mv.visitIntInsn(SIPUSH, values.length);
		mv.visitIntInsn(NEWARRAY, T_LONG);
		for(int i=0; i<values.length; i++) {
			mv.visitInsn(DUP);
			mv.visitIntInsn(SIPUSH, i);			
			mv.visitLdcInsn(values[i]);
			mv.visitInsn(LASTORE);
		}
	}
	
	private void pushPrimitiveDoubleArray(double[] values) {
		mv.visitIntInsn(SIPUSH, values.length);
		mv.visitIntInsn(NEWARRAY, T_DOUBLE);
		for(int i=0; i<values.length; i++) {
			mv.visitInsn(DUP);
			mv.visitIntInsn(SIPUSH, i);			
			mv.visitLdcInsn(values[i]);
			mv.visitInsn(DASTORE);
		}
	}
	
	private void pushPrimitiveFloatArray(float[] values) {
		mv.visitIntInsn(SIPUSH, values.length);
		mv.visitIntInsn(NEWARRAY, T_FLOAT);
		for(int i=0; i<values.length; i++) {
			mv.visitInsn(DUP);
			mv.visitIntInsn(SIPUSH, i);			
			mv.visitLdcInsn(values[i]);
			mv.visitInsn(FASTORE);
		}
	}
	
	private void pushPrimitiveBooleanArray(boolean[] values) {
		mv.visitIntInsn(SIPUSH, values.length);
		mv.visitIntInsn(NEWARRAY, T_BOOLEAN);
		for(int i=0; i<values.length; i++) {
			mv.visitInsn(DUP);
			mv.visitIntInsn(SIPUSH, i);
			int opCode = values[i] ? ICONST_1 : ICONST_0;
			mv.visitInsn(opCode);
			mv.visitInsn(BASTORE);
		}
	}

	private void pushObjectArray(Object[] values) {
		Class<?> componentType = values.getClass().getComponentType();
		String internalName = Type.getInternalName(componentType);
		mv.visitIntInsn(SIPUSH, values.length);
		mv.visitTypeInsn(ANEWARRAY, internalName);
		for(int i=0; i<values.length; i++) {
			mv.visitInsn(DUP);
			mv.visitIntInsn(SIPUSH, i);
			pushStringOrPrimitiveWrapper(values[i]);
			mv.visitInsn(AASTORE);
		}
	}
	
	private void createObjectGetter(Method m) {
		Object value = getStringOrPrimitiveWrapper(m);
		pushStringOrPrimitiveWrapper(value);
		mv.visitInsn(ARETURN);		
	}
	
	private void createPrimitiveByteGetter(Method m) {
		byte value;
		String stringValue = getStringPropValue(m);
		try {
			value = Byte.parseByte(String.valueOf(stringValue));
		} catch(NumberFormatException e) {
			throw new RapidPropsException("Invalid value for '" + getPropertyName(m) + "'", e);
		}
		mv.visitIntInsn(BIPUSH, value);
		mv.visitInsn(IRETURN);
	}

	private void createPrimitiveShortGetter(Method m) {
		short value;
		String stringValue = getStringPropValue(m);
		try {
			value = Short.parseShort(String.valueOf(stringValue));
		} catch(NumberFormatException e) {
			throw new RapidPropsException("Invalid value for '" + getPropertyName(m) + "'", e);
		}
		mv.visitIntInsn(SIPUSH, value);
		mv.visitInsn(IRETURN);
	}

	private void createPrimitiveIntGetter(Method m) {
		int value;
		String stringValue = getStringPropValue(m);
		try {
			value = Integer.parseInt(String.valueOf(stringValue));
		} catch(NumberFormatException e) {
			throw new RapidPropsException("Invalid value for '" + getPropertyName(m) + "'", e);
		}
		mv.visitLdcInsn(new Integer(value));
		mv.visitInsn(IRETURN);
	}

	private void createPrimitiveLongGetter(Method m) {
		long value;
		String stringValue = getStringPropValue(m);
		try {
			value = Long.parseLong(String.valueOf(stringValue));
		} catch(NumberFormatException e) {
			throw new RapidPropsException("Invalid value for '" + getPropertyName(m) + "'", e);
		}
		mv.visitLdcInsn(new Long(value));
		mv.visitInsn(LRETURN);
	}

	private void createPrimitiveDoubleGetter(Method m) {
		double value;
		String stringValue = getStringPropValue(m);
		try {
			value = Double.parseDouble(String.valueOf(stringValue));
		} catch(NumberFormatException e) {
			throw new RapidPropsException("Invalid value for '" + getPropertyName(m) + "'", e);
		}
		mv.visitLdcInsn(new Double(value));
		mv.visitInsn(DRETURN);
	}

	private void createPrimitiveFloatGetter(Method m) {
		float value;
		String stringValue = getStringPropValue(m);
		try {
			value = Float.parseFloat(String.valueOf(stringValue));
		} catch(NumberFormatException e) {
			throw new RapidPropsException("Invalid value for '" + getPropertyName(m) + "'", e);
		}
		mv.visitLdcInsn(new Float(value));
		mv.visitInsn(FRETURN);
	}

	private void createPrimitiveBooleanGetter(Method m) {
		boolean value;
		String stringValue = getStringPropValue(m);
		value = Boolean.parseBoolean(String.valueOf(stringValue));
		mv.visitLdcInsn(new Boolean(value));
		mv.visitInsn(IRETURN);
	}


	
	
	private static String getPropertyName(Method m) {
		Prop prop = m.getAnnotation(Prop.class);
		return getPropertyKey(m, prop);
	}
	
	protected static String getPropertyKey(Method m, Prop prop) {
		String propKey = "";
		if(prop != null) {
			propKey = prop.key();
		}
		if(propKey.isEmpty()) {
			String methodName = m.getName();
			String prefix = m.getReturnType().equals(boolean.class) ? "is" : "get";
			String errMsg = "Method " + methodName + " does not follow the java beans convention.";
			if(!methodName.startsWith(prefix)) throw new RapidPropsException(errMsg);
			String tmpName = methodName.substring(prefix.length());
			if(tmpName.isEmpty()) throw new RapidPropsException(errMsg);
			if(Character.isLowerCase(tmpName.charAt(0))) throw new RapidPropsException(errMsg);
			propKey = Character.toLowerCase(tmpName.charAt(0)) + tmpName.substring(1);
		}		
		return propKey;
	}
	
	private static String getDefaultValue(Prop prop) {
		String defaultValue = PropConstants.NOT_AVAILABLE;
		if(prop != null) {
			defaultValue = prop.defaultValue();
		}
		return defaultValue;
	}
	
	private static String getDefaultStringValue(String propKey, Prop prop) {
		String defaultValue = getDefaultValue(prop);
		if(defaultValue == PropConstants.NOT_AVAILABLE) {
			throw new RapidPropsException("No value provided for '" + propKey + "'");
		}
		if(Prop.NULL.equals(defaultValue)) {
			defaultValue = null;
		}
		return defaultValue;
	}

	public static Class<?> getParameterizedClass(Method m, int parameterIndex) {
		java.lang.reflect.Type type = m.getGenericReturnType();
		if(!(type instanceof ParameterizedType)) {
			throw new RapidPropsException(m.getName() + ": raw types are not supported.");
		} else {
			ParameterizedType parameterizedType = (ParameterizedType) type;			
			java.lang.reflect.Type[] parameterTypes = parameterizedType.getActualTypeArguments();
			if(parameterIndex >= parameterTypes.length) {
				throw new RapidPropsException(m.getName() + ": cannot retrieve generic parameter #" + parameterIndex + " from " + type);
			}
			java.lang.reflect.Type parameterType = parameterTypes[parameterIndex];
			if(parameterType instanceof WildcardType) {
				throw new RapidPropsException(m.getName() + ": wildcard types are not supported.");
			} else if(parameterType instanceof Class) {
				return (Class<?>)parameterType;
			} else if(parameterType instanceof GenericArrayType) {
				GenericArrayType genericArrayType = (GenericArrayType) parameterType;
				java.lang.reflect.Type genericComponentType = genericArrayType.getGenericComponentType();
				if(!(genericComponentType instanceof Class)) {
					throw new RapidPropsException(m.getName() + ": unsupported array type: " + genericArrayType);
				} else {
					return Array.newInstance((Class<?>)genericComponentType, 0).getClass();
				}
			} else {
				throw new RapidPropsException(m.getName() + ": unsupported parameter type: " + parameterType);
			}
		}
	}
}
